@@ -13,7 +13,7 @@
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	  $Id: pgtclId.c,v 1.36 2005/04/14 23:54:11 schwarzkopf Exp $
+ *	  $Id$
  *
  *-------------------------------------------------------------------------
  */
@@ -25,10 +25,6 @@
 #include "pgtclCmds.h"
 #include "pgtclId.h"
 
-#ifndef CONST84
-#     define CONST84
-#endif
-
 static int
 PgEndCopy(Pg_ConnectionId * connid, int *errorCodePtr)
 {
@@ -148,17 +144,23 @@ PgGetHandleProc(ClientData instanceData, int direction,
 }
 
 Tcl_ChannelType Pg_ConnType = {
-	"pgsql",					/* channel type */
-	NULL,						/* blockmodeproc */
-	PgDelConnectionId,			/* closeproc */
-	PgInputProc,				/* inputproc */
-	PgOutputProc,				/* outputproc */
-	NULL,						/* SeekProc, Not used */
-	NULL,						/* SetOptionProc, Not used */
-	NULL,						/* GetOptionProc, Not used */
-	PgWatchProc,				/* WatchProc, must be defined */
-	PgGetHandleProc,			/* GetHandleProc, must be defined */
-	NULL						/* Close2Proc, Not used */
+    "pgsql",             /* channel type */
+    NULL,                /* blockmodeproc */
+    PgDelConnectionId,   /* closeproc */
+    PgInputProc,         /* inputproc */
+    PgOutputProc,        /* outputproc */
+    NULL,                /* SeekProc, Not used */
+    NULL,                /* SetOptionProc, Not used */
+    NULL,                /* GetOptionProc, Not used */
+    PgWatchProc,         /* WatchProc, must be defined */
+    PgGetHandleProc,     /* GetHandleProc, must be defined */
+    NULL,                /* Close2Proc, Not used */
+    NULL,                /* blockModeProc, Not used */
+    NULL,                /* flushProc */
+    NULL,                /* handlerProc */
+    NULL,                /* wideSeekProc */
+    NULL,                /* threadActionProc */
+    NULL                 /* truncateProc */
 };
 
 /*
@@ -167,11 +169,11 @@ Tcl_ChannelType Pg_ConnType = {
 int
 PgSetConnectionId(Tcl_Interp *interp, PGconn *conn, char *chandle)
 {
-	Tcl_Channel conn_chan;
-        Tcl_Obj     *nsstr;
+	Tcl_Channel     conn_chan;
+        Tcl_Obj         *nsstr;
 	Pg_ConnectionId *connid;
-	int			i;
-        CONST   char      *ns = "";
+	int             i;
+        CONST char      *ns = "";
 
 	connid = (Pg_ConnectionId *) ckalloc(sizeof(Pg_ConnectionId));
 	connid->conn = conn;
@@ -183,6 +185,9 @@ PgSetConnectionId(Tcl_Interp *interp, PGconn *conn, char *chandle)
 	connid->res_copyStatus = RES_COPY_NONE;
 	connid->results = (PGresult **)ckalloc(sizeof(PGresult *) * RES_START);
 	connid->resultids = (Pg_resultid **)ckalloc(sizeof(Pg_resultid *) * RES_START);
+        connid->callbackPtr = (Tcl_Obj *) NULL;
+        connid->callbackInterp = (Tcl_Interp *) NULL;
+
 
 	for (i = 0; i < RES_START; i++)
 	{
@@ -193,6 +198,8 @@ PgSetConnectionId(Tcl_Interp *interp, PGconn *conn, char *chandle)
 	connid->notify_list = NULL;
 	connid->notifier_running = 0;
 	connid->interp = interp;
+	connid->nullValueString = NULL;
+	connid->sql_count = 0;
 
         nsstr = Tcl_NewStringObj("if {[namespace current] != \"::\"} {set k [namespace current]::}", -1);
 
@@ -209,14 +216,14 @@ PgSetConnectionId(Tcl_Interp *interp, PGconn *conn, char *chandle)
 
         if (chandle == NULL)
         {
-	    sprintf(connid->id, "%spgsql%d", ns, PQsocket(conn));
+	    sprintf(connid->id, "%spgsql%p", ns, connid);
         }
         else
         {
 	    sprintf(connid->id, "%s%s", ns, chandle);
         }
 
-    conn_chan = Tcl_GetChannel(interp, connid->id, 0);
+        conn_chan = Tcl_GetChannel(interp, connid->id, 0);
 
 	if (conn_chan != NULL)
 	{
@@ -243,7 +250,7 @@ PgSetConnectionId(Tcl_Interp *interp, PGconn *conn, char *chandle)
 	Tcl_SetResult(interp, connid->id, TCL_VOLATILE);
 	Tcl_RegisterChannel(interp, conn_chan);
 
-        connid->cmd_token=Tcl_CreateObjCommand(interp, connid->id, PgConnCmd, (ClientData) connid, PgDelCmdHandle);
+    connid->cmd_token=Tcl_CreateObjCommand(interp, connid->id, PgConnCmd, (ClientData) connid, PgDelCmdHandle);
 
     return 1;
 }
@@ -266,25 +273,36 @@ PgConnCmd(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
 {
     int             optIndex;
     int             objvxi;
+    int             idx = 1;
     char            *arg;
     Tcl_Obj         *objvx[25];
     Tcl_CmdInfo     info;
     Pg_ConnectionId *connid;
+    int             returnCode = TCL_ERROR;
 
-    static CONST84 char *options[] = {
-        "disconnect", "exec", "sqlexec", "execute", "select", "listen",
-        "on_connection_loss", "lo_creat", "lo_open", "lo_close", 
-        "lo_read", "lo_write", "lo_lseek", "lo_tell", "lo_unlink",
-        "lo_import", "lo_export", "sendquery", "exec_prepared", 
-        "sendquery_prepared",  (char *)NULL
+    static const char *options[] = {
+        "quote", "escape_bytea", "unescape_bytea", "disconnect", "exec", 
+	"sqlexec", "execute", "select", 
+	"listen", "on_connection_loss", "lo_creat", "lo_open", "lo_close", 
+        "lo_read", "lo_write", "lo_lseek", "lo_tell", "lo_truncate", 
+	"lo_unlink", "lo_import", "lo_export", "sendquery", "exec_prepared", 
+        "sendquery_prepared",  "null_value_string", "version", 
+        "protocol", "param", "backendpid", "socket", 
+	"conndefaults",  "set_single_row_mode", "is_busy", "blocking",
+	"cancel_request", (char *)NULL
     };
 
     enum options
     {
-        DISCONNECT,EXEC, SQLEXEC, EXECUTE, SELECT, LISTEN, 
-        ON_CONNECTION_LOSS, LO_CREAT, LO_OPEN, LO_CLOSE, LO_READ, 
-        LO_WRITE, LO_LSEEK, LO_TELL, LO_UNLINK, LO_IMPORT, LO_EXPORT, 
-        SENDQUERY, EXEC_PREPARED, SENDQUERY_PREPARED
+        QUOTE, ESCAPE_BYTEA, UNESCAPE_BYTEA, DISCONNECT, EXEC, 
+	SQLEXEC, EXECUTE, SELECT, 
+	LISTEN, ON_CONNECTION_LOSS, LO_CREAT, LO_OPEN, LO_CLOSE, 
+	LO_READ, LO_WRITE, LO_LSEEK, LO_TELL, LO_TRUNCATE, LO_UNLINK, 
+	LO_IMPORT, LO_EXPORT, SENDQUERY, EXEC_PREPARED, 
+	SENDQUERY_PREPARED, NULL_VALUE_STRING, VERSION, 
+	PROTOCOL, PARAM, BACKENDPID, SOCKET,
+	CONNDEFAULTS, SET_SINGLE_ROW_MODE, ISBUSY, BLOCKING,
+	CANCELREQUEST
     };
 
     if (objc == 1 || objc > 25)
@@ -305,26 +323,82 @@ PgConnCmd(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
     objvx[0] = objv[1];
     objvx[1] = objv[0];
 
-    if (Tcl_GetCommandInfo(interp, Tcl_GetStringFromObj(objvx[1], NULL), &info) == 0)
+    if (Tcl_GetCommandInfo(interp, Tcl_GetString(objvx[1]), &info) == 0)
         return TCL_ERROR;
 
     connid = (Pg_ConnectionId *) info.objClientData;
 
 
-    objvx[1] = Tcl_NewStringObj(connid->id, -1);
     if (Tcl_GetIndexFromObj(interp, objv[1], options, "command", TCL_EXACT, &optIndex) != TCL_OK)
                     return TCL_ERROR;
 
+    /*
+     *  Need to test here, since EXECUTE and UNESCAPE_BYTEA branches do things
+     *  a little differently
+    if ((optIndex != EXECUTE) && (optIndex != UNESCAPE_BYTEA))
+    {
+        objvx[1] = Tcl_NewStringObj(connid->id, -1);
+    }
+*/
+
     switch ((enum options) optIndex)
     {
+	case QUOTE:
+	{
+	    /* error if only two args, we gotta pick it up here or
+	     * Pg_quote will happily quote and return the connection ID.
+	     */
+	    if (objc == 2)
+	    {
+		Tcl_WrongNumArgs(interp, 1, objv, "quote string");
+		return TCL_ERROR;
+	    }
+
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_quote(cData, interp, objc, objvx);
+            break;
+	}
+
+	case ESCAPE_BYTEA:
+	{
+	    /* error if only two args, we gotta pick it up here or
+	     * Pg_escapeBytea will happily quote and return the connection ID.
+	     */
+	    if (objc == 2)
+	    {
+		Tcl_WrongNumArgs(interp, 1, objv, "escape_bytea byteArray");
+		return TCL_ERROR;
+	    }
+
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_escapeBytea(cData, interp, objc, objvx);
+            break;
+	}
+
+	case UNESCAPE_BYTEA:
+	{
+	    if (objc != 3)
+	    {
+		Tcl_WrongNumArgs(interp, 1, objv, "unescape_bytea string");
+		return TCL_ERROR;
+	    }
+	    objvx[1] = objv[2];
+            returnCode = Pg_unescapeBytea(cData, interp, 2, objvx);
+	    return returnCode;
+	}
+
         case DISCONNECT:
         {
-            return Pg_disconnect(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_disconnect(cData, interp, objc, objvx);
+            break;
         }
         case EXEC:
         case SQLEXEC:
         {
-            return Pg_exec(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_exec(cData, interp, objc, objvx);
+			break;
         }
         case EXECUTE:
         {
@@ -334,11 +408,11 @@ PgConnCmd(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
              * before the connection handle -- arrggh
              */
             int num = 2;
-            arg = Tcl_GetStringFromObj(objvx[2], NULL);
+            arg = Tcl_GetString(objvx[2]);
             if (arg[0] == '-')
             {
                 /* see if there are 2 options on the command line */
-                arg = Tcl_GetStringFromObj(objvx[4], NULL);
+                arg = Tcl_GetString(objvx[4]);
                 if (arg[0] == '-')
                 {
                     num = 4;
@@ -351,85 +425,189 @@ PgConnCmd(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
                 objvx[objvxi++] = objv[0];
                 
             }
-            /* DEBUGGING
+            /*  DEBUGGING
             for (objvxi = 0; objvxi < objc; objvxi++)
             {
-                printf("ID: %d OBJV %s OBJVX %s\n", objvxi, 
-                     Tcl_GetStringFromObj(objv[objvxi], NULL), 
-                     Tcl_GetStringFromObj(objvx[objvxi], NULL));
+                     Tcl_GetString(objv[objvxi]), 
+                     Tcl_GetString(objvx[objvxi]));
             }
             */
                
-            return Pg_execute(cData, interp, objc, objvx);
+            idx += num;
+            objvx[idx] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_execute(cData, interp, objc, objvx);
+			break;
         }
         case SELECT:
         {
-            return Pg_select(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_select(cData, interp, objc, objvx);
+			break;
         }
         case LISTEN:
         {
-            return Pg_listen(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_listen(cData, interp, objc, objvx);
+			break;
         }
         case ON_CONNECTION_LOSS:
         {
-            return Pg_listen(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_listen(cData, interp, objc, objvx);
+			break;
         }
         case LO_CREAT:
         {
-            return Pg_lo_creat(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_lo_creat(cData, interp, objc, objvx);
+			break;
         }
         case LO_OPEN:
         {
-            return Pg_lo_open(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_lo_open(cData, interp, objc, objvx);
+			break;
         }
         case LO_CLOSE:
         {
-            return Pg_lo_close(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_lo_close(cData, interp, objc, objvx);
+			break;
         }
         case LO_READ:
         {
-            return Pg_lo_read(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_lo_read(cData, interp, objc, objvx);
+			break;
         }
         case LO_WRITE:
         {
-            return Pg_lo_write(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_lo_write(cData, interp, objc, objvx);
+			break;
         }
         case LO_LSEEK:
         {
-            return Pg_lo_lseek(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_lo_lseek(cData, interp, objc, objvx);
+			break;
         }
         case LO_TELL:
         {
-            return Pg_lo_tell(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_lo_tell(cData, interp, objc, objvx);
+			break;
+        }
+        case LO_TRUNCATE:
+        {
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_lo_truncate(cData, interp, objc, objvx);
+			break;
         }
         case LO_UNLINK:
         {
-            return Pg_lo_unlink(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_lo_unlink(cData, interp, objc, objvx);
+			break;
         }
         case LO_IMPORT:
         {
-            return Pg_lo_import(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_lo_import(cData, interp, objc, objvx);
+			break;
         }
         case LO_EXPORT:
         {
-            return Pg_lo_export(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_lo_export(cData, interp, objc, objvx);
+			break;
         }
         case SENDQUERY:
         {
-            return Pg_sendquery(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_sendquery(cData, interp, objc, objvx);
+			break;
         }
         case EXEC_PREPARED:
         {
-            return Pg_exec_prepared(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_exec_prepared(cData, interp, objc, objvx);
+			break;
         }
         case SENDQUERY_PREPARED:
         {
-            return Pg_sendquery_prepared(cData, interp, objc, objvx);
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_sendquery_prepared(cData, interp, objc, objvx);
+			break;
         }
-    }
+        case NULL_VALUE_STRING:
+        {
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_null_value_string(cData, interp, objc, objvx);
+			break;
+        }
+        case PROTOCOL:
+        case BACKENDPID:
+        case SOCKET:
+        case VERSION:
+        {
+            
+            objc = 3;
+            objvx[2] = objv[0];
+            objvx[1] = objv[1];
+            idx++;
+            objvx[idx] = Tcl_NewStringObj(connid->id, -1);
+            returnCode= Pg_dbinfo(cData, interp, objc, objvx);
+	    break;
+        }
+        case PARAM:
+        {
+            
+            objc = 4;
+            objvx[2] = objv[0];
+            objvx[1] = objv[1];
+            objvx[3] = objv[2];
+            idx++;
+            objvx[idx] = Tcl_NewStringObj(connid->id, -1);
+            returnCode= Pg_dbinfo(cData, interp, objc, objvx);
+	    break;
+        }
+	case CONNDEFAULTS:
+	{
+            returnCode = Pg_conndefaults(cData, interp, 1, objvx);
+	    break;
+	}
+	
+	case SET_SINGLE_ROW_MODE:
+	{
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_set_single_row_mode(cData, interp, objc, objvx);
+	    break;
+	}
+	
+	case ISBUSY:
+	{
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_isbusy(cData, interp, objc, objvx);
+	    break;
+	}
+	
+	case BLOCKING:
+	{
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_blocking(cData, interp, objc, objvx);
+	    break;
+	}
 
-	/* It should not be possible to get here */
-    return TCL_ERROR;
+	case CANCELREQUEST:
+	{
+            objvx[1] = Tcl_NewStringObj(connid->id, -1);
+            returnCode = Pg_cancelrequest(cData, interp, objc, objvx);
+	    break;
+	}
+    }
+	Tcl_DecrRefCount(objvx[idx]);
+	return returnCode;
 }
 
 /* 
@@ -477,7 +655,7 @@ PgResultCmd(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[
  * Get back the connection from the Id
  */
 PGconn *
-PgGetConnectionId(Tcl_Interp *interp, CONST84 char *id, Pg_ConnectionId ** connid_p)
+PgGetConnectionId(Tcl_Interp *interp, const char *id, Pg_ConnectionId ** connid_p)
 {
 	Tcl_Channel conn_chan;
 	Pg_ConnectionId *connid;
@@ -516,18 +694,27 @@ PgDelConnectionId(DRIVER_DEL_PROTO)
 	int			i;
         Pg_resultid    *resultid;
 
+
 	connid = (Pg_ConnectionId *) cData;
 
 	for (i = 0; i < connid->res_max; i++)
 	{
-            resultid = connid->resultids[i];
 	    if (connid->results[i])
-            {
-		PQclear(connid->results[i]);
+		{
+			PQclear(connid->results[i]);
 
-            Tcl_DecrRefCount(resultid->str);
-            }
-            ckfree((void *)resultid);
+
+            resultid = connid->resultids[i];
+
+			if (resultid != NULL) {
+				Tcl_DecrRefCount(resultid->str);
+
+				if ((resultid->nullValueString != NULL) && (resultid->nullValueString != connid->nullValueString))
+					ckfree (resultid->nullValueString);
+
+				ckfree((void *)resultid);
+			}
+		}
 	}
 	
 	ckfree((void *)connid->results);
@@ -540,7 +727,7 @@ PgDelConnectionId(DRIVER_DEL_PROTO)
 		for (entry = Tcl_FirstHashEntry(&notifies->notify_hash, &hsearch);
 			 entry != NULL;
 			 entry = Tcl_NextHashEntry(&hsearch))
-			ckfree((char *)Tcl_GetHashValue(entry));
+			ckfree((void *)Tcl_GetHashValue(entry));
 		Tcl_DeleteHashTable(&notifies->notify_hash);
 		if (notifies->conn_loss_cmd)
 			ckfree((void *) notifies->conn_loss_cmd);
@@ -561,6 +748,9 @@ PgDelConnectionId(DRIVER_DEL_PROTO)
 	PQfinish(connid->conn);
 	connid->conn = NULL;
 
+	if (connid->nullValueString != NULL)
+		ckfree(connid->nullValueString);
+
 	/*
 	 * Kill the notifier channel, too.	We must not do this until after
 	 * we've closed the libpq connection, because Tcl will try to close
@@ -574,12 +764,25 @@ PgDelConnectionId(DRIVER_DEL_PROTO)
 	 * Note we are not leaking a socket, since libpq closed that already.
 	 */
 
-#if TCL_MAJOR_VERSION >= 8
 	if (connid->notifier_channel != NULL && interp != NULL)
         {
 		Tcl_UnregisterChannel(NULL, connid->notifier_channel);
          }
-#endif
+
+       /*
+        * Clear any async result callback, if present.
+        */
+
+        if (connid->callbackPtr)    {
+           Tcl_DecrRefCount(connid->callbackPtr);
+           connid->callbackPtr = NULL;
+        }
+
+        if (connid->callbackInterp) {
+           Tcl_Release((ClientData) connid->interp);
+           connid->callbackInterp = NULL;
+        }
+
 
 	/*
 	 * We must use Tcl_EventuallyFree because we don't want the connid
@@ -613,7 +816,7 @@ PgDelConnectionId(DRIVER_DEL_PROTO)
  */
 
 int
-PgSetResultId(Tcl_Interp *interp, CONST84 char *connid_c, PGresult *res)
+PgSetResultId(Tcl_Interp *interp, const char *connid_c, PGresult *res, int *idPtr)
 {
     Tcl_Channel     conn_chan;
     Pg_ConnectionId *connid;
@@ -677,25 +880,34 @@ PgSetResultId(Tcl_Interp *interp, CONST84 char *connid_c, PGresult *res)
 	}
 
     }
+
     connid->results[resid] = res;
+
     sprintf(buf, "%s.%d", connid_c, resid);
     cmd = Tcl_NewStringObj(buf, -1);
+
     resultid = (Pg_resultid *) ckalloc(sizeof(Pg_resultid));
+
     resultid->interp = interp;
     resultid->id     = resid;
     resultid->str = Tcl_NewStringObj(buf, -1);
     resultid->cmd_token = Tcl_CreateObjCommand(interp, buf, 
         PgResultCmd, (ClientData) resultid, PgDelResultHandle);
+	resultid->connid = connid;
+	resultid->nullValueString = connid->nullValueString;
+
     connid->resultids[resid] = resultid;
 
     Tcl_SetObjResult(interp, cmd);
 
-    return resid;
+    *idPtr = resid;
+
+    return TCL_OK;
 }
 
 
 static int
-getresid(Tcl_Interp *interp, CONST84 char *id, Pg_ConnectionId ** connid_p)
+getresid(Tcl_Interp *interp, const char *id, Pg_ConnectionId ** connid_p)
 {
 	Tcl_Channel conn_chan;
 	char	   *mark;
@@ -740,7 +952,7 @@ getresid(Tcl_Interp *interp, CONST84 char *id, Pg_ConnectionId ** connid_p)
  * Get back the result pointer from the Id
  */
 PGresult *
-PgGetResultId(Tcl_Interp *interp, CONST84 char *id)
+PgGetResultId(Tcl_Interp *interp, const char *id, Pg_resultid **resultidPtr)
 {
 	Pg_ConnectionId *connid;
 	int			resid;
@@ -750,6 +962,10 @@ PgGetResultId(Tcl_Interp *interp, CONST84 char *id)
 	resid = getresid(interp, id, &connid);
 	if (resid == -1)
 		return NULL;
+
+	if (resultidPtr != NULL) {
+		*resultidPtr = connid->resultids[resid];
+	}
 	return connid->results[resid];
 }
 
@@ -758,9 +974,10 @@ PgGetResultId(Tcl_Interp *interp, CONST84 char *id)
  * Remove a result Id from the hash tables
  */
 void
-PgDelResultId(Tcl_Interp *interp, CONST84 char *id)
+PgDelResultId(Tcl_Interp *interp, const char *id)
 {
 	Pg_ConnectionId *connid;
+	Pg_resultid     *resultid;
 	int			resid;
 
 	resid = getresid(interp, id, &connid);
@@ -769,8 +986,14 @@ PgDelResultId(Tcl_Interp *interp, CONST84 char *id)
 
 	connid->results[resid] = 0;
 
-Tcl_DecrRefCount((Tcl_Obj *)connid->resultids[resid]->str);
-ckfree((void *)connid->resultids[resid]);
+	resultid = connid->resultids[resid];
+
+	Tcl_DecrRefCount((Tcl_Obj *)resultid->str);
+
+	if ((resultid->nullValueString != NULL) && (resultid->nullValueString != connid->nullValueString))
+		ckfree (resultid->nullValueString);
+
+	ckfree((void *)resultid);
 	connid->resultids[resid] = 0;
 }
 
@@ -779,7 +1002,7 @@ ckfree((void *)connid->resultids[resid]);
  * Get the connection Id from the result Id
  */
 int
-PgGetConnByResultId(Tcl_Interp *interp, CONST84 char *resid_c)
+PgGetConnByResultId(Tcl_Interp *interp, const char *resid_c)
 {
 	char	   *mark;
 	Tcl_Channel conn_chan;
@@ -1081,6 +1304,38 @@ AllNotifyEventDeleteProc(Tcl_Event *evPtr, ClientData clientData)
 	}
 	return 0;
 }
+static int
+Pg_Result_EventProc(Tcl_Event *evPtr, int flags)
+{
+    NotifyEvent *event = (NotifyEvent *) evPtr;
+
+    /* Results can only come from file events. */
+    if (!(flags & TCL_FILE_EVENTS))
+       return 0;
+
+    /* If connection's been closed, just forget the whole thing. */
+    if (event->connid) {
+       Pg_ConnectionId *connid = event->connid;
+       Tcl_Obj *callbackPtr = connid->callbackPtr;
+       Tcl_Interp *interp = connid->callbackInterp;
+
+       /* Clear the result callback for this connection, so that the callback
+        * script may safely establish a new one. */
+
+       connid->callbackPtr = NULL;
+       connid->callbackInterp = NULL;
+
+       if (callbackPtr || interp) {
+           if (TCL_OK != Tcl_EvalObjEx(interp, callbackPtr, TCL_EVAL_GLOBAL)) {
+               Tcl_BackgroundError(interp);
+           }
+           Tcl_DecrRefCount(callbackPtr);
+           Tcl_Release((ClientData) interp);
+       }
+    }
+    /* never deliver this event twice */
+    return 1;
+}
 
 /*
  * File handler callback: called when Tcl has detected read-ready on socket.
@@ -1102,6 +1357,25 @@ Pg_Notify_FileHandler(ClientData clientData, int mask)
 	{
 		/* Transfer notify events from libpq to Tcl event queue. */
 		PgNotifyTransferEvents(connid);
+
+               /* If the connection is still alive, and if there is a
+                * callback for results, check if a result is ready. If it is,
+                * transfer the event to the Tcl event queue.
+                */
+
+               if ((PQsocket(connid->conn) >= 0)
+                       && connid->callbackPtr
+                       && !PQisBusy(connid->conn)) {
+
+                   NotifyEvent *event = (NotifyEvent *) ckalloc(sizeof(NotifyEvent));
+
+                   event->header.proc = Pg_Result_EventProc;
+                   event->notify = NULL;
+                   event->connid = connid;
+                   Tcl_QueueEvent((Tcl_Event *) event, TCL_QUEUE_TAIL);
+                }
+
+
 	}
 	else
 	{
@@ -1217,9 +1491,9 @@ PgDelResultHandle(ClientData cData)
     Pg_resultid    *resultid = (Pg_resultid *) cData;
     char           *resstr;
 
-    resstr = Tcl_GetStringFromObj(resultid->str, NULL);
+    resstr = Tcl_GetString(resultid->str);
     
-    result = PgGetResultId(resultid->interp, resstr);
+    result = PgGetResultId(resultid->interp, resstr, NULL);
 
     PgDelResultId(resultid->interp, resstr);
     PQclear(result);
